import os 

import FreeCAD as App
from . import settings
import Part

from kipy.board_types import Footprint3DModel, BoardPolygon, BoardSegment, PadStackShape
from kipy.geometry import PolygonWithHoles, PolyLine, PolyLineNode, Vector2
from kipy.proto.common.types import KiCadObjectType
from kipy.util.board_layer import BoardLayer

from .bases import BaseObject, BaseViewProvider
from . import parts

class BoardObject(BaseObject):
    TYPE = 'KiConnect::Board'

    def __init__(self, feature):
        super(BoardObject, self).__init__(feature)

        self.kicad_board = None
        self.substrate_body = None
        self.substrate_sketch = None
        self.via_sketch = None

        feature.addProperty('App::PropertyPlacement', 'BoardOffset', 'KiConnect', 'Internal offset for zeroing out Footprint offset', hidden=True, read_only=True)

    def onDocumentRestored(self, feature):
        if self.kicad_board is None:
            self.kicad_board = self.get_api().kicad.get_board()

        if self.substrate_body is None:
            self.extrude_substrate(feature)
            self.sketch_outline(feature)

        if len(self.kicad_board.get_vias()) > 0 and self.via_sketch is None:
            self.setup_vias()
            self.pocket_vias()

    def extrude_substrate(self, feature):
        self.substrate_sketch = App.ActiveDocument.addObject('Sketcher::SketchObject', 'Sketch')
        self.substrate_sketch.Visibility = False

        self.substrate_body = App.ActiveDocument.addObject('PartDesign::Body', 'Substrate')
        self.feature.addObject(self.substrate_body)

        pad = self.substrate_body.newObject('PartDesign::Pad', 'Outline')

        pad.Profile = self.substrate_sketch
        pad.Length = 1.6
        pad.Midplane = True

        self.substrate_body.addObject(self.substrate_sketch)

    def get_boardpoly(self):
        board = self.kicad_board

        # find polygons of Edge Cuts
        edge_cuts = [ edge for edge in  board.get_shapes() if edge.layer == BoardLayer.BL_Edge_Cuts ]
        polygons = [ edge for edge in edge_cuts if isinstance(edge, BoardPolygon) ]

        # XXX only single board supported at the moment
        return polygons[0]

    def pocket_vias(self):
        board = self.kicad_board

        self.via_sketch.Geometry = []

        for via in board.get_vias():
            stack = via.padstack
            f_cu = stack.copper_layer(BoardLayer.BL_F_Cu)
                
            self.via_sketch.addGeometry(Part.Circle((App.Vector(via.position.x, -via.position.y) / 1000000.0) - self.feature.BoardOffset.Base, App.Vector(0, 0, 1), via.drill_diameter / 1000000.0 / 2))

        for pad in board.get_pads():
            copper_layer = pad.padstack.copper_layer(BoardLayer.BL_F_Cu)
            drill = pad.padstack.drill

            if copper_layer.shape is not PadStackShape.PSS_CIRCLE: continue

            center = (App.Vector(pad.position.x, -pad.position.y) / 1000000.0) - self.feature.BoardOffset.Base

            if drill.diameter.x == drill.diameter.y:
                rad = drill.diameter.x / 1000000.0 / 2

                self.via_sketch.addGeometry(Part.Circle(center, App.Vector(0, 0, 1), rad))

    def setup_vias(self):
        self.via_sketch = App.ActiveDocument.addObject('Sketcher::SketchObject')
        self.via_sketch.Visibility = False
        self.substrate_body.addObject(self.via_sketch)
        via_pocket = self.substrate_body.newObject('PartDesign::Pocket', 'Vias')
        via_pocket.Profile = self.via_sketch
        via_pocket.Midplane = True
        via_pocket.Type = 1

    def sketch_outline(self, feature):
        boardpoly = self.get_boardpoly()
        poly = boardpoly.polygons[0]

        # track KIID
        # TODO FeaturePython objects interacting with kipy objects should inherit common parent class
        self.feature.PolygonId = boardpoly.id.value
        
        # this offset centers the board to 0,0
        bb = boardpoly.bounding_box()
        self.feature.BoardOffset.Base = (App.Vector(bb.pos.x, -bb.pos.y) + App.Vector(bb.size.x, -bb.size.y) / 2) / 1000000.0

        begin = None
        start = None
        
        # reset Sketch Geometry
        self.substrate_sketch.Geometry = []

        # sketch outline
        for segment in poly.outline:
            if not start:
                start = (App.Vector(segment.point.x, -segment.point.y)) / 1000000.0 - self.feature.BoardOffset.Base
                
                # needs to remain to connect the last segment
                begin = start

                continue

            end = (App.Vector(segment.point.x, -segment.point.y)) / 1000000.0 - self.feature.BoardOffset.Base

            self.substrate_sketch.addGeometry(
                Part.LineSegment(start, end)
            )

            start = end

        # make final connection
        self.substrate_sketch.addGeometry(
            Part.LineSegment(start, begin)
        )

    def sync(self):
        board = self.kicad_board

        commit = board.begin_commit()

        edge_cuts = [ edge for edge in  board.get_shapes() if edge.layer == BoardLayer.BL_Edge_Cuts ]
        polygons = [ edge for edge in edge_cuts if isinstance(edge, BoardPolygon) ]

        # XXX only single board supported at the moment
        boardpoly = polygons[0]
        poly = boardpoly.polygons[0]
        poly.outline.clear()

        geom = self.feature.getObject('Substrate').getObject('Sketch').Geometry
        segments = [l for l in geom if isinstance(l, Part.LineSegment)]

        for line in segments:
            start = (line.StartPoint + self.feature.BoardOffset.Base) * 1000000
            start.y = -start.y

            poly.outline.append(PolyLineNode.from_point(Vector2.from_xy(int(start.x), int(start.y))))

        board.update_items(boardpoly)
        board.push_commit(commit, "Updated board outline")

        # if this is done, probably need to clear selection at the start
        # maybe save the selection and restore if not empty, else select the poly as below
        #board.add_to_selection(boardpoly)

    def __getstate__(self):
        return None

class BoardViewProvider(BaseViewProvider):
    TYPE = 'KiConnect::Board'
    ICON = 'board.svg'

class Board:
    def __init__(self, kicad_board, parent):
        feature = App.ActiveDocument.addObject('App::DocumentObjectGroupPython', 'Board')
        parent.addObject(feature)
        self.feature = feature

        BoardObject(feature)
        BoardViewProvider(feature.ViewObject)

        # TODO move into BoardObject
        feature.addProperty('App::PropertyString', 'Doc', 'KiConnect', 'Doc in project to sync with', read_only=True).Doc = kicad_board.name
        feature.addProperty('App::PropertyString', 'PolygonId', 'KiConnect', 'Polygon ID for the original outline', hidden=True, read_only=True)

        parts.makeParts(feature)

